import json as json_lib

class Parser:
    def __init__(self):
        pass

    def parse_json(self,payload,model,payload_list=False):
        try:
            json = json_lib.loads(payload)
        except Exception as e:
            raise BundestagsAPyException(f'Failed to parse JSON payload: {e}')
        if 'cursor' in json:
            cursor = json['cursor']
        else:
            cursor = None
        if 'numFound' in json:
            numFound = json['numFound']
        else:
            numFound=None
        if payload_list:
            result = model.parse_list(json)
        else:
            result = model.parse(json)   
        if cursor:
            if numFound or numFound==0:
                return result, cursor, numFound
            else:
                return result, cursor
        else:
            return result


import requests
import logging
import sys
from platform import python_version
log = logging.getLogger(__name__)
import time
from BundestagsAPy.errors import (
    BadRequest, Forbidden, HTTPException, NotFound, TooManyRequests,
    BundestagsAPyException, Unauthorized)
from BundestagsAPy.models import (Fundstelle,Aktivitaet,Drucksache,DrucksacheText,
                    Person,Plenarprotokoll,PlenarprotokollText,Vorgang,Vorgangsposition)
from BundestagsAPy.parsers import Parser

class Client:
    def __init__(self,api_key=None,retry_count=0,retry_delay=0,retry_errors=None,
                timeout=60,user_agent=None):
        if api_key==None:
            raise BundestagsAPyException("API Key is required. See https://dip.bundestag.de/%C3%BCber-dip/hilfe/api#content on how to get a valid API Key.")
        self.api_key=api_key
        self.retry_count=retry_count
        self.retry_errors=retry_errors
        self.retry_delay=retry_delay
        if user_agent is None:
            user_agent = (
                f"Python/{python_version()} "
                f"Requests/{requests.__version__} "
                )
        self.user_agent=user_agent
        self.session = requests.Session()
        self.parser=Parser()
    
    def call(
        self,endpoint,model,id=None,endpoint_parameters=(),
        headers=None,params=None,payload_list=False,**kwargs):
        if headers is None:
            headers = {}
        headers["User-Agent"]=self.user_agent
        headers['Authorization']=f"ApiKey {self.api_key}"

        #build URL
        if id is not None:
            url = f"https://search.dip.bundestag.de/api/v1/{endpoint}/{id}"
        else:
            url = f"https://search.dip.bundestag.de/api/v1/{endpoint}"
        
        if params is None:
            params = {}
            params['format']='json'
    
        for k, arg in kwargs.items():
            if arg is None:
                continue
            if k not in endpoint_parameters:
                log.warning(f'Unexpected parameter: {k}')
            if k in ['format','cursor']:
                params[k] = str(arg)
            elif k =='start_date':
                params['f.datum.start'] = str(arg)
            elif k =='end_date':
                params['f.datum.end'] = str(arg)
            else:
                params[f"f.{k}"] = str(arg)
        log.debug("PARAMS: %r", params)
        
        try:
            #attempt request until it is successful or the maximum
            #retries are reached
            retries_performed = 0
            while retries_performed<=self.retry_count:
                try:
                    response = self.session.request(
                        'GET',url,params=params,headers=headers)
                except Exception as e:
                    raise BundestagsAPyException(f"Failed to send request: {e}").with_traceback(sys.exc_info()[2])
                
                if response.status_code==200:
                    break
            
                retry_delay = self.retry_delay
            
                if not self.retry_errors or response.status_code not in self.retry_errors:
                    break
            
                # Sleep before retrying request again
                time.sleep(retry_delay)
                retries_performed += 1
        
            #Exceptions for Errors
            self.last_response = response
            
            if response.status_code == 400:
                raise BadRequest(response)
            if response.status_code == 401:
                raise Unauthorized(response)
            if response.status_code == 403:
                raise Forbidden(response)
            if response.status_code == 404:
                raise NotFound(response)
            if response.status_code == 429:
                raise TooManyRequests(response)
            if response.status_code and not response.status_code==200:
                raise HTTPException(response)
        
            # Parse the response payload
            parser=self.parser
            result = parser.parse_json(
                response.text, model=model,payload_list=payload_list)
            
            return result
        finally:
            self.session.close()
    
    def get_data(self,endpoint,model,max_results=50,endpoint_parameters=(),**kwargs):
        #run the actual request and iterate through cursor
        #if the request is only for one file
        if 'id' in kwargs:
            id = kwargs['id']
            if isinstance(id,list):
                if len(id)==1:
                    return self.call(endpoint,model=model,id=id[0],
                                     endpoint_parameters=endpoint_parameters)
            elif isinstance(id,int):
                return self.call(endpoint,model=model,id=id,
                                 endpoint_parameters=endpoint_parameters)
        
        #all other cases
        results_collected = 0
        total_results=1
        results = []
        while results_collected < max_results or not max_results:
            result = self.call(endpoint,model=model,
                               endpoint_parameters=endpoint_parameters,payload_list=True,
                              **kwargs)
            if isinstance(result,tuple):
                results = results + result[0]
                results_collected+=len(result[0])
                kwargs['cursor']=result[1]
                total_results=result[2]
                if results_collected>=total_results:
                    break
            else:
                results = results + result[0]
                break     
        return results
    
    #Aktivitaet requests
    def bt_aktivitaet(self,max_results = 50, **kwargs):
        """
        bt_aktivitaet(id,start_date,end_date,drucksache,
            plenarprotokoll,zuordnung,format,max_results,cursor)
            
        Returns the at least the first 50 results.
        max_results may be False, in which case all results will be returned
        """
        #define endpoint parameters
        endpoint_parameters=(
                'id', 'start_date', 'end_date', 'drucksache', 'plenarprotokoll', 'zuordnung',
                'format', 'cursor')
        
        model =Aktivitaet()   
        return self.get_data('aktivitaet',model,max_results,endpoint_parameters,**kwargs)
    
    #Drucksache requests
    def bt_drucksache(self,max_results = 50, **kwargs):
        """
        bt_drucksache(id,start_date,end_date,zuordnung,format,max_results,cursor)
            
        Returns the at least the first 50 results.
        max_results may be False, in which case all results will be returned
        """
        #define endpoint parameters
        endpoint_parameters=(
                'id', 'start_date', 'end_date', 'zuordnung',
                'format', 'cursor')
        
        model =Drucksache()   
        return self.get_data('drucksache',model,max_results,endpoint_parameters,**kwargs)
    
    #DrucksacheText requests
    def bt_drucksache_text(self,max_results = 50, **kwargs):
        """
        bt_drucksache_text(id,start_date,end_date,format,max_results,cursor)
            
        Returns the at least the first 50 results.
        max_results may be False, in which case all results will be returned
        """
        #define endpoint parameters
        endpoint_parameters=(
                'id', 'start_date', 'end_date','format', 'cursor')
        
        model =DrucksacheText()   
        return self.get_data('drucksache-text',model,max_results,endpoint_parameters,**kwargs)
    
    #Person requests
    def bt_person(self,max_results = 50, **kwargs):
        """
        bt_person(id,start_date,end_date,format,max_results,cursor)
            
        Returns the at least the first 50 results.
        max_results may be False, in which case all results will be returned
        """
        #define endpoint parameters
        endpoint_parameters=(
                'id', 'start_date', 'end_date','format', 'cursor')
        
        model = Person()   
        return self.get_data('person',model,max_results,endpoint_parameters,**kwargs)
    
    #Plenarprotokoll requests
    def bt_plenarprotokoll(self,max_results = 50, **kwargs):
        """
        bt_plenarprotokoll(id,start_date,end_date,zuordnung,format,max_results,cursor)
            
        Returns the at least the first 50 results.
        max_results may be False, in which case all results will be returned
        """
        #define endpoint parameters
        endpoint_parameters=(
                'id', 'start_date', 'end_date','zuordnung','format', 'cursor')
        
        model = Plenarprotokoll()   
        return self.get_data('plenarprotokoll',model,max_results,endpoint_parameters,**kwargs)

    #PlenarprotokollText requests
    def bt_plenarprotokoll_text(self,max_results = 50, **kwargs):
        """
        bt_plenarprotokoll_text(id,start_date,end_date,format,max_results,cursor)
            
        Returns the at least the first 50 results.
        max_results may be False, in which case all results will be returned
        """
        #define endpoint parameters
        endpoint_parameters=(
                'id', 'start_date', 'end_date','format', 'cursor')
        
        model = PlenarprotokollText()   
        return self.get_data('plenarprotokoll-text',model,max_results,endpoint_parameters,**kwargs)
    
    #Vorgang requests
    def bt_vorgang(self,max_results = 50, **kwargs):
        """
        bt_vorgang(id,start_date,end_date,drucksache,
            plenarprotokoll,format,max_results,cursor)
            
        Returns the at least the first 50 results.
        max_results may be False, in which case all results will be returned
        """
        #define endpoint parameters
        endpoint_parameters=(
                'id', 'start_date', 'end_date', 'drucksache', 'plenarprotokoll',
                'format', 'cursor')
        
        model =Vorgang()   
        return self.get_data('vorgang',model,max_results,endpoint_parameters,**kwargs)
    
    #Vorgang requests
    def bt_vorgangsposition(self,max_results = 50, **kwargs):
        """
        bt_vorgangsposition(id,start_date,end_date,drucksache,
            plenarprotokoll,vorgang,zuordnung,format,max_results,cursor,)
            
        Returns the at least the first 50 results.
        max_results may be False, in which case all results will be returned
        """
        #define endpoint parameters
        endpoint_parameters=(
                'id', 'start_date', 'end_date', 'drucksache', 'plenarprotokoll','vorgang','zuordnung',
                'format', 'cursor')
        
        model =Vorgangsposition()   
        return self.get_data('vorgangsposition',model,max_results,endpoint_parameters,**kwargs)
    


import json

class BundestagsAPyException(Exception):
    """Base exception for BundestagsAPy"""
    pass

class HTTPException(BundestagsAPyException):
    """Exception raised when an HTTP request fails"""

    def __init__(self, response):
        self.response = response

        self.api_errors = []
        self.api_codes = []
        self.api_messages = []

        try:
            response_json = response.json()
        except json.JSONDecodeError:
            super().__init__(f"{response.status_code} {response.reason}")
        else:
            errors = response_json.get("errors", [])
            if "error" in response_json:
                errors.append(response_json["error"])
            error_text = ""
            for error in errors:
                self.api_errors.append(error)
                if "code" in error:
                    self.api_codes.append(error["code"])
                if "message" in error:
                    self.api_messages.append(error["message"])
                if "code" in error and "message" in error:
                    error_text += f"\n{error['code']} - {error['message']}"
                elif "message" in error:
                    error_text += '\n' + error["message"]
            super().__init__(
                f"{response.status_code} {response.reason}{error_text}"
            )

class BadRequest(HTTPException):
    """Exception raised for a 400 HTTP status code"""
    pass


class Unauthorized(HTTPException):
    """Exception raised for a 401 HTTP status code"""
    pass


class Forbidden(HTTPException):
    """Exception raised for a 403 HTTP status code"""
    pass


class NotFound(HTTPException):
    """Exception raised for a 404 HTTP status code"""
    pass


class TooManyRequests(HTTPException):
    """Exception raised for a 429 HTTP status code"""
    pass


class Model:
    def __init__(self):
        pass
    @classmethod
    def parse(cls,json):
        pass
    @classmethod
    def parse_list(cls,json):
        pass

class Fundstelle(Model):
    @classmethod
    def parse(cls,json):
        fundstelle = cls()
        setattr(fundstelle, '_json', json)
        for k, v in json.items():
            setattr(fundstelle, k, v)
        return fundstelle
    
class Aktivitaet(Model):
    @classmethod
    def parse(cls,json):
        aktivitaet = cls()
        setattr(aktivitaet, '_json', json)
        for k, v in json.items():
            if k == 'fundstelle':
                fundstelle = Fundstelle.parse(v)
                setattr(aktivitaet,'fundstelle',fundstelle)
            else:
                setattr(aktivitaet, k, v)
        return aktivitaet
    @classmethod
    def parse_list(cls,json_list):
        item_list = json_list['documents']
        results = []
        for item in item_list:
            results.append(cls.parse(item))
        return results

class Drucksache(Model):
    @classmethod
    def parse(cls,json):
        ds = cls()
        setattr(ds, '_json', json)
        for k, v in json.items():
            if k == 'fundstelle':
                fundstelle = Fundstelle.parse(v)
                setattr(ds,'fundstelle',fundstelle)
            else:
                setattr(ds, k, v)
        return ds
    @classmethod
    def parse_list(cls,json_list):
        item_list = json_list['documents']
        results = []
        for item in item_list:
            results.append(cls.parse(item))
        return results

class DrucksacheText(Model):
    @classmethod
    def parse(cls,json):
        dst = cls()
        setattr(dst, '_json', json)
        for k, v in json.items():
            if k == 'fundstelle':
                fundstelle = Fundstelle.parse(v)
                setattr(dst,'fundstelle',fundstelle)
            else:
                setattr(dst, k, v)
        return dst
    @classmethod
    def parse_list(cls,json_list):
        item_list = json_list['documents']
        results = []
        for item in item_list:
            results.append(cls.parse(item))
        return results

class Person(Model):
    @classmethod
    def parse(cls,json):
        ps = cls()
        setattr(ps, '_json', json)
        for k, v in json.items():
            setattr(ps, k, v)
        return ps
    @classmethod
    def parse_list(cls,json_list):
        item_list = json_list['documents']
        results = []
        for item in item_list:
            results.append(cls.parse(item))
        return results

class Plenarprotokoll(Model):
    @classmethod
    def parse(cls,json):
        pp = cls()
        setattr(pp, '_json', json)
        for k, v in json.items():
            if k == 'fundstelle':
                fundstelle = Fundstelle.parse(v)
                setattr(pp,'fundstelle',fundstelle)
            else:
                setattr(pp, k, v)
        return pp
    @classmethod
    def parse_list(cls,json_list):
        item_list = json_list['documents']
        results = []
        for item in item_list:
            results.append(cls.parse(item))
        return results

class PlenarprotokollText(Model):
    @classmethod
    def parse(cls,json):
        ppt = cls()
        setattr(ppt, '_json', json)
        for k, v in json.items():
            if k == 'fundstelle':
                fundstelle = Fundstelle.parse(v)
                setattr(ppt,'fundstelle',fundstelle)
            else:
                setattr(ppt, k, v)
        return ppt
    @classmethod
    def parse_list(cls,json_list):
        item_list = json_list['documents']
        results = []
        for item in item_list:
            results.append(cls.parse(item))
        return results

class Vorgang(Model):
    @classmethod
    def parse(cls,json):
        vg = cls()
        setattr(vg, '_json', json)
        for k, v in json.items():
            setattr(vg, k, v)
        return vg
    @classmethod
    def parse_list(cls,json_list):
        item_list = json_list['documents']
        results = []
        for item in item_list:
            results.append(cls.parse(item))
        return results

class Vorgangsposition(Model):
    @classmethod
    def parse(cls,json):
        vgp = cls()
        setattr(vgp, '_json', json)
        for k, v in json.items():
            if k == 'fundstelle':
                fundstelle = Fundstelle.parse(v)
                setattr(vgp,'fundstelle',fundstelle)
            else:
                setattr(vgp, k, v)
        return vgp
    @classmethod
    def parse_list(cls,json_list):
        item_list = json_list['documents']
        results = []
        for item in item_list:
            results.append(cls.parse(item))
        return results


# BundestagsAPy

"""
BundetagsAPy Bundestags DIP API Library
"""

__version__ = '1.2-alpha'
__author__ = 'Paul Bose'
__license__ = 'MIT'

from BundestagsAPy.client import Client
from BundestagsAPy.parsers import Parser
from BundestagsAPy.models import (Fundstelle,Aktivitaet,Drucksache,DrucksacheText,
                    Person,Plenarprotokoll,PlenarprotokollText,Vorgang,Vorgangsposition)
from BundestagsAPy.errors import (
    BadRequest, Forbidden, HTTPException, NotFound, TooManyRequests,
    BundestagsAPyException, Unauthorized)


from setuptools import setup, find_packages
import re

VERSION_FILE = "BundestagsAPy/__init__.py"
with open(VERSION_FILE) as version_file:
    match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                      version_file.read(), re.MULTILINE)

if match:
    version = match.group(1)
else:
    raise RuntimeError(f"Unable to find version string in {VERSION_FILE}.")

with open("README.md") as readme_file:
    long_description = readme_file.read()

# Setting up
setup(
        name="BundestagsAPy", 
        version=version,
        author="Paul Bose",
        author_email="<bose@ese.eur.nl>",
        description='Python Wrapper for the Bundestags API',
        long_description=long_description,
        long_description_content_type="text/markdown",
        packages=find_packages(),
        install_requires=["requests>=2.26.0,<3",],
    
        keywords=['Bundestag','API'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 3.7",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ]
)